package DNS::Resolver;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

DNS::Resolver - Short module description

=head1 SYNOPSIS

 use DNS::Resolver;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Storable 'dclone';
use Socket;
use IO::Select;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use Net::DNS;

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Data::Cache::Memory;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant TIMEOUT => 1;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#
my $SINGLETON_INSTANCE;

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item singleton() [PUBLIC,STATIC]

  Usage     : my $instance = DNS::Resolver->singleton()
  Purpose   : Create/return reference to the singleton instance
  Arguments : Same as the the new() method
  Returns   : Same as the the new() method
  Throws    : Same as the the new() method
  Comments  : Internally uses new() method
  See Also  : new() method

=cut

sub singleton
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    unless ($SINGLETON_INSTANCE) {
        $SINGLETON_INSTANCE = $class->new(@_);
    }
    return $SINGLETON_INSTANCE;
}

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = DNS::Resolver->new();
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : DNS::Resolver reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating DNS.Resolver instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying DNS.Resolver instance...
}

=item resolve_ip($) [PUBLIC]

  Usage     : my $hostname = $resolver->resolve_ip($ip);
  Purpose   : Resolve given IPv4 address to hostname
  Arguments : MIXED $ip - IPv4 address [MANDATORY]
  Returns   : STRING $hostname
  Throws    : Croaks, if invoked on object

=cut

sub resolve_ip($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ip,) = @_;

    # Resolve IP address to hostname
    my $hostname = $self->{CACHE}->fetch($ip);
    unless (defined $hostname) {
        $hostname = $self->_resolve($ip, 'PTR');
        $self->{CACHE}->cache($ip, $hostname);
    }
    return $hostname;
}

=item resolve_host($) [PUBLIC]

  Usage     : my $ip = $resolver->resolve_host($host);
  Purpose   : Resolve given hostname to IPv4 address
  Arguments : STRING $host - hostname to resolve [MANDATORY]
  Returns   : STRING $ip
  Throws    : Croaks, if invoked on object

=cut

sub resolve_host($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($hostname,) = @_;

    # Resolve hostname to IP address
    my $ip = $self->{CACHE}->fetch($hostname);
    unless (defined $ip) {
        $ip = $self->_resolve($hostname, "A");
        $self->{CACHE}->cache($hostname, $ip);
    }
    return $ip;
}

=item resolve(;$$) [PUBLIC]

  Usage     : my ($ip, $host) = $resolver->resolve($ip, $host);
  Purpose   : Resolve the missing argument (IP or HOSTNAME) and return complete pair
  Arguments : MIXED $ip - IPv4 adress [OPTIONAL]
              STRING $host - hostname [OPTIONAL]
  Returns   : LIST containing STRING $ip and STRING $host
  Throws    : Croaks, if invoked on object

=cut

sub resolve(;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ip, $hostname) = @_;

    # Resolve IP address to hostname
    if ($ip and not $hostname) {
        return ($ip, $self->resolve_ip($ip));
    }
    # Resolve hostname to IP address
    elsif ($hostname and not $ip) {
        return ($self->resolve_host($hostname), $hostname);
    }
    # Best case scenario: nothing to do
    return ($ip, $hostname);
}

=item stats() [PUBLIC]

  Usage     : my $stats = $resolver->stats();
  Purpose   : Get resolver statistics
  Arguments : NONE
  Returns   : HASH REFERENCE $stats
  Throws    : Croaks, if invoked on object

=cut

sub stats()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $s = dclone($self->{CACHE}->stats());
    foreach my $k (keys %{$self->{STATS}}) {
        $s->{$k} = $self->{STATS}->{$k};
    }
    return $s;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new DNS::Resolver instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : DNS::Resolver $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Template.Module instance...
    my ($timeout,) = @_;

    $self->{TO}       = ($timeout)?$timeout:TIMEOUT();
    $self->{RESOLVER} = Net::DNS::Resolver->new();
    $self->{SELECT}   = IO::Select->new();
    $self->{CACHE}    = Data::Cache::Memory->new();
    $self->{STATS}    = {};
    return $self;
}

# _resolve($$) [PROTECTED]
#
#  Usage     : my $result = $self->_resolve($what, $qtype);
#  Purpose   : Resolve given object (IP or hostname)
#  Arguments : STRING $what - object to be resolved [MANDATORY]
#  Returns   : STRING $result
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _resolve($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($what, $qtype) = @_;

    my $sock = $self->{RESOLVER}->bgsend($what);
    $self->{SELECT}->add($sock);
    my $result = '';

    my @ready = $self->{SELECT}->can_read($self->{TO});
    if (@ready) {
        foreach my $s (@ready) {
            if ($s == $sock) {
                my $reply = $self->{RESOLVER}->bgread($sock);

                if ($reply) {
                    foreach my $rr ($reply->answer) {
                        if ($rr->type eq 'PTR') {
                            $self->{STATS}->{'resolved_host'}++;
                            $result = $rr->ptrdname;
                            goto REND;
                        }
                        elsif ($rr->type eq 'A') {
                            $self->{STATS}->{'resolved_ip'}++;
                            $result = $rr->address;
                            goto REND;
                        }
                    }
                } else {
                    Log::Loger->warn(__PACKAGE__, "[DNS] DNS query failed: ".$reply->errorstring);
                    $self->{STATS}->{'failure'}++;
                }
            }
            $s = undef;
        }
    } else {
        Log::Loger->debug(__PACKAGE__, "[DNS] DNS query timeouted after '$self->{TO}' seconds");
        $self->{STATS}->{'timeout'}++;
    }

    REND:
    $self->{SELECT}->remove($sock);
    $sock = undef;

    $self->{STATS}->{'unresolved'}++ unless $result;
    return $result;
}

=pod

=back

=cut

1;
