package Data::Query::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Query::Mongo - Execute query on objects stored within MongoDB database

=head1 SYNOPSIS

 use Data::Query::Mongo;

 # Instantinate the mongo query object
 my $query = Data::Query::Mongo->new(Data::Convertor::Mongo->new(), Mentat::Storage::Mongo->new());

 # Query the MongoDB database for messages that pass the given filter
 my $messages = $query->query($filter);

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Query(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Data::Filter::Traverser;
use Data::Query;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Data::Filter::Traverser', 'Data::Query');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item generate_query($) [PUBLIC]

  Usage     : my $query = $instance->generate_query($filter);
  Purpose   : Generate query for the MongoDB database
  Arguments : Data::Filter $filter - query filter tree
  Returns   : HASH REFERENCE containing MongoDB query
  Throws    : Croaks, if not invoked on object

=cut

sub generate_query($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;
    confess "Filter must be given as argument" unless $filter and blessed($filter);

    # Build the query hash
    my $query = $filter->traverse($self);

    print STDERR "GENERATED QUERY: " . Dumper($query) if $DEVEL;
    return $query;
}

=item query($;$$$$$) [PUBLIC]

  Usage     : my $result = $instance->query($filter, $selection, $sort, $limit, $skip, $quick);
  Purpose   : Query the MongoDB database for generic objects
  Arguments : Data::Filter $filter - query filter tree
              HASH REFERENCE $selection - field selection specification
              HASH REFERENCE $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
              BOOL $quick - perform quick query
  Returns   : ARRAY REFERENCE containing Mentat messages in raw database format
  Throws    : Croaks, if not invoked on object

=cut

sub query($;$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $selection, $sort, $limit, $skip, $quick) = @_;

    # Build the query hash
    my $query = $self->generate_query($filter);

    # Query the database
    return $self->storage->find($query, $selection, $sort, $limit, $skip, $quick);
}

=item query_json($;$$$$$) [PUBLIC]

  Usage     : my $mentat_messages = $instance->query_json($filter, $selection, $sort, $limit, $skip, $quick)
  Purpose   : Query the MongoDB database for generic objects in native JSON
  Arguments : (JSON STRING|HASH REFERENCE) $filter - query filter tree
              (JSON STRING|HASH REFERENCE) $selection - field selection specification
              (JSON STRING|HASH REFERENCE) $sort - sorting specification
              INTEGER $limit - maximal number of records returned
              INTEGER $skip - number of records to skip from beginning
              BOOL $quick - perform quick query
  Returns   : ARRAY REFERENCE containing Mentat::Message objects
  Throws    : Croaks, if not invoked on object

=cut

sub query_json($;$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $selection, $sort, $limit, $skip, $quick) = @_;

    my ($flt_j, $sel_j, $srt_j);
    if (ref($filter) eq 'HASH')    { $flt_j = $filter; }
    else                           { $flt_j = ($filter) ? decode_json($filter) : undef; }
    if (ref($selection) eq 'HASH') { $sel_j = $selection; }
    else                           { $sel_j = ($selection) ? decode_json($selection) : undef; }
    if (ref($sort) eq 'HASH')      { $srt_j = $sort; }
    else                           { $srt_j = ($sort) ? decode_json($sort) : undef; }

    # Query the database
    return $self->storage->find($flt_j, $sel_j, $srt_j, $limit, $skip, $quick);
}

#-------------------------------------------------------------------------------

=item FILTER CALLBACKS [PUBLIC]

  Usage     : return $traverser->$name(@args);
  Purpose   : Rule tree traversal callbacks, for each rule class one callback method
  Arguments : Depend on particular callback
  Returns   : Any result from callback will be passed to the parent node callback as argument
  Throws    : Croaks, if invoked on object
  Comments  : These methods are called from Data::Filter::traverse method
              during filter tree traversal. Arguments are dependent on
  See Also  :

=cut

sub constant($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, @args) = @_;

    return $filter->value();
}
sub list($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, @args) = @_;

    return $filter->value();
}
sub existence($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, @args) = @_;

    my $path = $filter->value(); $path =~ s|/|.|g;
    return { $path => { '$exists' => 1 } };
}
sub variable($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, @args) = @_;

    my $path = $filter->value(); $path =~ s|/|.|g;
    return $path;
}
# ---
sub operationUnary($$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->operationUnary_NOT(@_);
}
sub operationBinary($$$$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, @args) = @_;
    $operation = lc($operation);

    my $type = $self->operation_type($operation);

    return $self->operationBinary_lgcl($filter, $operation, @args) if $type eq Data::Filter::Traverser::OP_LOGIC;
    return $self->operationBinary_comp($filter, $operation, @args) if $type eq Data::Filter::Traverser::OP_COMP;
    return $self->operationBinary_calc($filter, $operation, @args) if $type eq Data::Filter::Traverser::OP_CALC;
    die "Invalid binary operation '$operation'\n";
}

#-------------------------------------------------------------------------------

sub operationUnary_NOT($$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $subfilter, $operand, @args) = @_;

    return { '$not' => $operand };
}

#-------------------------------------------------------------------------------

sub operationBinary_IN($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $l_subfilter, $r_subfilter, $l_operand, $r_operand, @args) = @_;

    my %mapping = ('in' => '$in');

    my $value;
    my $path;
    my $i=0;
    foreach (@$r_operand) {
        my $conversion = $self->{CONVERTOR}->convert_value($l_operand, $_, $operation);
        $path = $conversion->{PATH} if ($i == 0);
        push (@$value, $conversion->{VALUE});

        $i++;
    }
    $l_operand = $path;
    $r_operand = $value;

    return { $l_operand => { $mapping{$operation} => $r_operand} } if $mapping{$operation};
}

sub operationBinary_LIKE($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $l_subfilter, $r_subfilter, $l_operand, $r_operand, @args) = @_;

    return { $l_operand => qr/$r_operand/ };
}

#-------------------------------------------------------------------------------

sub operationBinary_lgcl($$$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $l_subfilter, $r_subfilter, $l_operand, $r_operand, @args) = @_;

    my %mapping = ('and' => '$and', 'or' => '$or', 'xor' => '$xor');

    return { $mapping{$operation} => [$l_operand, $r_operand] } if $mapping{$operation};
    die "Invalid logical operation '$operation'\n";
}

sub operationBinary_comp($$$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $l_subfilter, $r_subfilter, $l_operand, $r_operand, @args) = @_;

    my %mapping = ('eq' => '', 'ne' => '$ne', 'lt' => '$lt', 'gt' => '$gt', 'le' => '$lte', 'ge' => '$gte');
    my $operator = $mapping{$operation};

    if (defined $operator) {
        my $conversion = $self->convertor->convert_value($l_operand, $r_operand, $operator);

        $l_operand = $conversion->{PATH};
        $r_operand = $conversion->{VALUE};

        return (($conversion->{REPLACE} == 1) ? $r_operand : { $l_operand => $r_operand }) unless $operator;
        return  ($conversion->{REPLACE} == 1) ? $r_operand : { $l_operand => { $operator => $r_operand }};
    }

    # Operations IN and LIKE are specific
    return $self->operationBinary_IN(@_)   if $operation eq 'in';
    return $self->operationBinary_LIKE(@_) if $operation eq 'like';
    die "Invalid comparison operation '$operation'\n";
}

sub operationBinary_calc($$$$$;@)
{
    my $self = shift;
    #confess "Instance method not invoked on object instance" unless blessed($self);
    #my ($filter, $operation, $l_subfilter, $r_subfilter, $l_operand, $r_operand, @args) = @_;

    die "This feature is not yet implemented and supported\n";
}

#-------------------------------------------------------------------------------

=item storage(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if not invoked on object

=cut

sub storage(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($storage,) = @_;

    # Setter mode
    if ($storage) {
        confess "Valid Mentat::Storage::Mongo must be given as argument" unless blessed($storage) and $storage->isa("Mentat::Storage::Mongo");
        $self->{STORAGE} = $storage;
    }
    return $self->{STORAGE};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Query::Mongo instance
#  Arguments :
#  Returns   : Data::Query::Mongo reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %args = @_;
    confess "Data convertor must be given as argument" unless $args{convertor};
    confess "Data storage must be given as argument" unless $args{storage};

    $self->convertor($args{convertor});
    $self->storage($args{storage});

    return $self;
}

=pod

=back

=cut

1;
