package Mentat::Message::Correlator::Context::Queue;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Correlator::Context::Queue - Queue implementation for message correlation contexts

=head1 SYNOPSIS

 use Mentat::Message::Correlator::Context::Queue;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Message::Correlator::Context::Queue;->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Message::Correlator::Context::Queue; reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Message.Correlator.Context.Queue instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Message.Correlator.Context.Queue instance...
}

#-------------------------------------------------------------------------------

=item enqueue($) [PUBLIC]

  Usage     : $instance->enqueue($context);
  Purpose   : Enqueue given context into context queue
  Arguments : STRING $context_id - context identifier [MANDATORY]
              Mentat::Message::Correlator::Context $context - context to be queued [MANDATORY]
  Returns   : Nothing
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub enqueue($)
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
    my ($context,) = @_;
}

=item dequeue() [PUBLIC]

  Usage     : my $context = $instance->dequeue();
  Purpose   : Fetch next timeouted context from the queue
  Arguments : None
  Returns   : Mentat::Message::Correlator::Context $context or UNDEF
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub dequeue()
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
}

#-------------------------------------------------------------------------------

=item update($) [PUBLIC]

  Usage     : my $error = $instance->update($context_id, $data);
  Purpose   : Update context with given identifier with given data
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : BOOL $result
  Throws    : Dies, if not invoked on object instance or if given invalid arguments
  Comment   : It is not allowed to update context ID or timeout

=cut

sub update($)
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
    my ($context_id, $data) = @_;
}

=item fetch($) [PUBLIC]

  Usage     : my $error = $instance->fetch($context_id);
  Purpose   : Fetch context with given identifier
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : Mentat::Message::Correlator::Context $context or UNDEF
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub fetch($)
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
    my ($context_id,) = @_;
}

#-------------------------------------------------------------------------------

=item count() [PUBLIC]

  Usage     : my $count = $instance->count();
  Purpose   : Get number of contexts in queue
  Arguments : None
  Returns   : INTEGER $count
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub count()
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
}

=item exists($) [PUBLIC]

  Usage     : my $result = $instance->exists($context_id);
  Purpose   : Check, if the context with given ID is already present in the queue
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : BOOL $result
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub exists($)
{
    my $self = shift;
    confess "Method needs to be implemented in subclass";
    my ($context_id,) = @_;
}

=item statistics() [PUBLIC]

  Usage     : my $statistics = $instance->statistics();
  Purpose   : Get processing statistics
  Arguments : None
  Returns   : HASH REFERENCE $statistics with various keys
  Throws    : Dies, if not invoked on object instance or if given invalid arguments
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub statistics()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{STATISTICS};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Correlator::Context::Queue; instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Message::Correlator::Context::Queue; $self reference
#  Throws    : Dies, if not invoked on object instance or if given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat.Message.Correlator.Context.Queue instance...
    my %config = @_;

    # Processing statistics (for debugging, development and optimization purposes)
    $self->{STATISTICS} = {};

    return $self;
}

=pod

=back

=cut

1;
